package meerkat.bulletinboard;

import com.google.common.util.concurrent.FutureCallback;
import com.google.protobuf.ByteString;
import meerkat.comm.CommunicationException;
import meerkat.protobuf.BulletinBoardAPI.*;
import meerkat.protobuf.Voting.*;

import java.util.List;

/**
 * Created by Arbel Deutsch Peled on 03-Mar-16.
 * This is a full-fledged implementation of a Bulletin Board Client
 * It provides asynchronous access to several remote servers, as well as a local cache
 * Read operations are performed on the local server
 * Batch reads are performed on the local server and, if they fail, also on the remote servers
 * Write operations are performed first on the local server and then on the remotes
 * After any read is carried out, a subscription is made for the specific query to make sure the local DB will be updated
 * The database also employs a synchronizer which makes sure local data is sent to the remote servers
 */
public class CachedBulletinBoardClient implements SubscriptionBulletinBoardClient {

    private final AsyncBulletinBoardClient localClient;
    private AsyncBulletinBoardClient remoteClient;
    private BulletinBoardSubscriber subscriber;

    private class SubscriptionStoreCallback implements FutureCallback<List<BulletinBoardMessage>>  {

        private final FutureCallback<?> callback;

        public SubscriptionStoreCallback(){
            callback = null;
        }

        public SubscriptionStoreCallback(FutureCallback<?> callback){
            this.callback = callback;
        }

        @Override
        public void onSuccess(List<BulletinBoardMessage> result) {
            for (BulletinBoardMessage msg : result) {
                try {
                    localClient.postMessage(msg);
                } catch (CommunicationException ignored) {
                    // TODO: log
                }
            }
        }

        @Override
        public void onFailure(Throwable t) {
            if (callback != null) {
                callback.onFailure(t);  // This is some hard error that cannot be dealt with
            }
        }

    }

    /**
     * Creates a Cached Client
     * Assumes all parameters are initialized
     * @param localClient is a Client for the local instance
     * @param remoteClient is a Client for the remote instance(s); Should have endless retries for post operations
     * @param subscriber is a subscription service to the remote instance(s)
     */
    public CachedBulletinBoardClient(AsyncBulletinBoardClient localClient,
                                     AsyncBulletinBoardClient remoteClient,
                                     BulletinBoardSubscriber subscriber) {

        this.localClient = localClient;
        this.remoteClient = remoteClient;
        this.subscriber = subscriber;

    }

    @Override
    public MessageID postMessage(final BulletinBoardMessage msg, final FutureCallback<Boolean> callback) {

        return localClient.postMessage(msg, new FutureCallback<Boolean>() {
            @Override
            public void onSuccess(Boolean result) {
                remoteClient.postMessage(msg, callback);
            }

            @Override
            public void onFailure(Throwable t) {
                if (callback != null)
                    callback.onFailure(t);
            }
        });

    }

    @Override
    public MessageID postBatch(final CompleteBatch completeBatch, final FutureCallback<Boolean> callback) {

        return localClient.postBatch(completeBatch, new FutureCallback<Boolean>() {
            @Override
            public void onSuccess(Boolean result) {
                remoteClient.postBatch(completeBatch, callback);
            }

            @Override
            public void onFailure(Throwable t) {
                if (callback != null)
                    callback.onFailure(t);
            }
        });

    }

    @Override
    public void beginBatch(final BeginBatchMessage beginBatchMessage, final FutureCallback<Boolean> callback) {

        localClient.beginBatch(beginBatchMessage, new FutureCallback<Boolean>() {
            @Override
            public void onSuccess(Boolean result) {
                remoteClient.beginBatch(beginBatchMessage, callback);
            }

            @Override
            public void onFailure(Throwable t) {
                if (callback != null)
                    callback.onFailure(t);
            }
        });

    }

    @Override
    public void postBatchData(final byte[] signerId, final int batchId, final List<BatchData> batchDataList,
                              final int startPosition, final FutureCallback<Boolean> callback) {

        localClient.postBatchData(signerId, batchId, batchDataList, startPosition, new FutureCallback<Boolean>() {
            @Override
            public void onSuccess(Boolean result) {
                remoteClient.postBatchData(signerId, batchId, batchDataList, startPosition, callback);
            }

            @Override
            public void onFailure(Throwable t) {
                if (callback != null)
                    callback.onFailure(t);
            }
        });

    }

    @Override
    public void postBatchData(final byte[] signerId, final int batchId, final List<BatchData> batchDataList, final FutureCallback<Boolean> callback) {

        localClient.postBatchData(signerId, batchId, batchDataList, new FutureCallback<Boolean>() {
            @Override
            public void onSuccess(Boolean result) {
                remoteClient.postBatchData(signerId, batchId, batchDataList, callback);
            }

            @Override
            public void onFailure(Throwable t) {
                if (callback != null)
                    callback.onFailure(t);
            }
        });

    }

    @Override
    public void postBatchData(final ByteString signerId, final int batchId, final List<BatchData> batchDataList,
                              final int startPosition, final FutureCallback<Boolean> callback) {

        localClient.postBatchData(signerId, batchId, batchDataList, startPosition, new FutureCallback<Boolean>() {
            @Override
            public void onSuccess(Boolean result) {
                remoteClient.postBatchData(signerId, batchId, batchDataList, startPosition, callback);
            }

            @Override
            public void onFailure(Throwable t) {
                if (callback != null)
                    callback.onFailure(t);
            }
        });

    }

    @Override
    public void postBatchData(final ByteString signerId, final int batchId, final List<BatchData> batchDataList, final FutureCallback<Boolean> callback) {

        localClient.postBatchData(signerId, batchId, batchDataList, new FutureCallback<Boolean>() {
            @Override
            public void onSuccess(Boolean result) {
                remoteClient.postBatchData(signerId, batchId, batchDataList, callback);
            }

            @Override
            public void onFailure(Throwable t) {
                if (callback != null)
                    callback.onFailure(t);
            }
        });

    }

    @Override
    public void closeBatch(final CloseBatchMessage closeBatchMessage, final FutureCallback<Boolean> callback) {

        localClient.closeBatch(closeBatchMessage, new FutureCallback<Boolean>() {
            @Override
            public void onSuccess(Boolean result) {
                remoteClient.closeBatch(closeBatchMessage, callback);
            }

            @Override
            public void onFailure(Throwable t) {
                if (callback != null)
                    callback.onFailure(t);
            }
        });

    }

    @Override
    public void getRedundancy(MessageID id, FutureCallback<Float> callback) {

        remoteClient.getRedundancy(id, callback);

    }

    @Override
    public void readMessages(MessageFilterList filterList, final FutureCallback<List<BulletinBoardMessage>> callback) {

        localClient.readMessages(filterList, callback);

        subscriber.subscribe(filterList, new SubscriptionStoreCallback(callback));

    }

    @Override
    public void readBatch(final BatchSpecificationMessage batchSpecificationMessage, final FutureCallback<CompleteBatch> callback) {

        localClient.readBatch(batchSpecificationMessage, new FutureCallback<CompleteBatch>() {
            @Override
            public void onSuccess(CompleteBatch result) {
                if (callback != null)
                    callback.onSuccess(result);     // Read from local client was successful
            }

            @Override
            public void onFailure(Throwable t) {

                // Read from local unsuccessful: try to read from remote

                remoteClient.readBatch(batchSpecificationMessage, new FutureCallback<CompleteBatch>() {

                    @Override
                    public void onSuccess(CompleteBatch result) {

                        // Read from remote was successful: store in local and return result

                        localClient.postBatch(result, new FutureCallback<Boolean>() {
                            @Override
                            public void onSuccess(Boolean result) {}
                            @Override
                            public void onFailure(Throwable t) {}
                        });

                        if (callback != null)
                            callback.onSuccess(result);

                    }

                    @Override
                    public void onFailure(Throwable t) {

                        // Read from remote was unsuccessful: report error
                        if (callback != null)
                            callback.onFailure(t);

                    }

                });

            }

        });

    }

    @Override
    public void querySync(SyncQuery syncQuery, FutureCallback<SyncQueryResponse> callback) {

        localClient.querySync(syncQuery, callback);

    }

    @Override
    /**
     * This is a stub method
     * All resources are assumed to be initialized
     */
    public void init(BulletinBoardClientParams clientParams) {}

    @Override
    public MessageID postMessage(BulletinBoardMessage msg) throws CommunicationException {
        localClient.postMessage(msg);
        return remoteClient.postMessage(msg);
    }

    @Override
    public float getRedundancy(MessageID id) {
        return remoteClient.getRedundancy(id);
    }

    @Override
    public List<BulletinBoardMessage> readMessages(MessageFilterList filterList) throws CommunicationException {
        subscriber.subscribe(filterList, new SubscriptionStoreCallback());
        return localClient.readMessages(filterList);
    }

    @Override
    public CompleteBatch readBatch(BatchSpecificationMessage batchSpecificationMessage) throws CommunicationException {
        return localClient.readBatch(batchSpecificationMessage);
    }

    @Override
    public SyncQuery generateSyncQuery(GenerateSyncQueryParams generateSyncQueryParams) throws CommunicationException {
        return localClient.generateSyncQuery(generateSyncQueryParams);
    }

    @Override
    public void close() {
        localClient.close();
        remoteClient.close();
    }

    @Override
    public void subscribe(MessageFilterList filterList, FutureCallback<List<BulletinBoardMessage>> callback) {
        subscriber.subscribe(filterList, callback);
    }

    @Override
    public void subscribe(MessageFilterList filterList, long startEntry, FutureCallback<List<BulletinBoardMessage>> callback) {
        subscriber.subscribe(filterList, startEntry, callback);
    }

}