package meerkat.bulletinboard;

import com.google.common.util.concurrent.FutureCallback;
import com.google.protobuf.ByteString;
import meerkat.comm.CommunicationException;
import meerkat.protobuf.BulletinBoardAPI.*;
import meerkat.util.BulletinBoardUtils;

import java.util.LinkedList;
import java.util.List;

/**
 * Created by Arbel on 13/04/2016.
 * Simple, straightforward implementation of the {@link BulletinBoardSynchronizer} interface
 */
public class SimpleBulletinBoardSynchronizer implements BulletinBoardSynchronizer {

    private DeletableSubscriptionBulletinBoardClient localClient;
    private AsyncBulletinBoardClient remoteClient;

    private volatile SyncStatus syncStatus;

    private List<FutureCallback<Integer>> messageCountCallbacks;
    private List<FutureCallback<SyncStatus>> syncStatusCallbacks;

    private static final MessageFilterList EMPTY_FILTER = MessageFilterList.getDefaultInstance();
    private static final int SLEEP_INTERVAL = 10000; // 10 Seconds

    private class SyncCallback implements FutureCallback<List<BulletinBoardMessage>> {

        @Override
        public void onSuccess(List<BulletinBoardMessage> result) {

            SyncStatus newStatus = SyncStatus.PENDING;

            if (result.size() == 0) {
                newStatus = SyncStatus.SYNCHRONIZED;
            }

            else{   // Upload messages

                for (BulletinBoardMessage message : result){

                    if (message.getMsg().getTagList().contains(BulletinBoardConstants.BATCH_TAG)){

                        // This is a batch message: need to upload batch data as well as the message itself
                        ByteString signerId = message.getSig(0).getSignerId();
                        long batchID = Long.parseLong(BulletinBoardUtils.findTagWithPrefix(message, BulletinBoardConstants.BATCH_ID_TAG_PREFIX));

                        BatchSpecificationMessage batchSpecificationMessage = BatchSpecificationMessage.newBuilder().build();

                        localClient.readBatch(batchSpecificationMessage, null);

                    }
                    else{

                        // This is a regular message: post it
                        try {
                            remoteClient.postMessage(message);
                        } catch (CommunicationException e) {
                            newStatus = SyncStatus.SERVER_ERROR;
                        }

                    }
                }

            }

            updateSyncStatus(newStatus);

        }

        @Override
        public void onFailure(Throwable t) {

            updateSyncStatus(SyncStatus.SERVER_ERROR);

        }

    }

    public SimpleBulletinBoardSynchronizer() {
        this.syncStatus = SyncStatus.STOPPED;
    }

    private synchronized void updateSyncStatus(SyncStatus newStatus) {

        if (newStatus != syncStatus){

            syncStatus = newStatus;

            for (FutureCallback<SyncStatus> callback : syncStatusCallbacks){
                if (callback != null)
                    callback.onSuccess(syncStatus);
            }

        }

    }

    @Override
    public void init(DeletableSubscriptionBulletinBoardClient localClient, AsyncBulletinBoardClient remoteClient) {

        updateSyncStatus(SyncStatus.STOPPED);

        this.localClient = localClient;
        this.remoteClient = remoteClient;

        messageCountCallbacks = new LinkedList<>();
        syncStatusCallbacks = new LinkedList<>();

    }

    @Override
    public SyncStatus getSyncStatus() {
        return syncStatus;
    }

    @Override
    public void subscribeToSyncStatus(FutureCallback<SyncStatus> callback) {
        syncStatusCallbacks.add(callback);
    }

    @Override
    public List<BulletinBoardMessage> getRemainingMessages()  throws CommunicationException{
        return localClient.readMessages(EMPTY_FILTER);
    }

    @Override
    public void getRemainingMessages(FutureCallback<List<BulletinBoardMessage>> callback) {
        localClient.readMessages(EMPTY_FILTER, callback);
    }

    @Override
    public long getRemainingMessagesCount() throws CommunicationException {
        return localClient.readMessages(EMPTY_FILTER).size();
    }

    @Override
    public void subscribeToRemainingMessagesCount(FutureCallback<Integer> callback) {
        messageCountCallbacks.add(callback);
    }

    @Override
    public void run() {

        if (syncStatus != SyncStatus.STOPPED) {

            updateSyncStatus(SyncStatus.PENDING);
            SyncCallback callback = new SyncCallback();

            while (syncStatus != SyncStatus.STOPPED) {

                try {

                    do {
                        localClient.readMessages(EMPTY_FILTER, callback);
                    } while (syncStatus == SyncStatus.PENDING);

                    synchronized (this) {
                        this.wait(SLEEP_INTERVAL);
                    }

                } catch (InterruptedException e) {
                    e.printStackTrace();
                }

            }

        }

    }

    @Override
    public void nudge() {
        synchronized (this) {
            this.notify();
        }
    }

    @Override
    public void stop() {

        updateSyncStatus(SyncStatus.STOPPED);

    }

}
